"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _babelTemplate = require("babel-template");

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

var _babelGenerator = require("babel-generator");

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

var _arrayFind = require("array-find");

var _arrayFind2 = _interopRequireDefault(_arrayFind);

var _path = require("path");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var LOG_CACHE = {};

/**
 * Convenience method to log additional work to do a single time per node type.
 *
 * @param {string} fnName function being called
 * @param {Node} node AST node
 * @param {Path} [path] optional AST Path for printing additional context.
 **/
/* eslint no-extra-parens: 0 */
var logOnce = function logOnce(fnName, node, path) {
  console.log("logOnce", node);
  var name = fnName + "-" + node.type;
  if (!LOG_CACHE[name]) {
    LOG_CACHE[name] = true;
    console.log("TODO(%s): handle %s.\n%s\n", fnName, node.type, (0, _babelGenerator2.default)(path ? path.node : node).code);
  }
};

/**
 * TemplateElement value nodes must be of the shape {raw: string; value: string}
 */
var templateElementValue = function templateElementValue(value) {
  return { raw: value, cooked: value };
};

exports.default = function (_ref) {
  var t = _ref.types;

  var ROOT_CSSNAMES_IDENTIFIER = "cssmodule";
  var BAIL_OUT = "__dontTransformMe";

  var matchesPatterns = function matchesPatterns(path, patterns) {
    return !!(0, _arrayFind2.default)(patterns, function (pattern) {
      return t.isIdentifier(path.node, { name: pattern }) || path.matchesPattern(pattern);
    });
  };

  /**
   * Updates a ConditionalExpression consequent or alternate node with the most
   * appropriate CSS Module lookup.
   *
   * @param {Path} path consequent or alternate node of a conditional expression
   * @param {Node<Identifier>} cssmodule cssmodule identifier
   */
  var replaceConditionalExpression = function replaceConditionalExpression(path, cssmodule) {
    return path.replaceWith(computeClassName(path, cssmodule)); // eslint-disable-line
  };

  /**
   * Generate the required TemplateElements for the following type of template:
   * `${ slot } ${ anotherslot }`
   */
  var spacedTemplateElements = function spacedTemplateElements(count) {
    return Array.apply(0, Array(count)).map(function (_, i) {
      return i === 0 || i === count - 1 ? t.templateElement(templateElementValue(""), i === count) : t.templateElement(templateElementValue(" "), false);
    });
  };

  var computeClassName = function computeClassName(value, cssmodule) {
    if (t.isStringLiteral(value)) {
      if (value.node.value === "") {
        return value.node;
      }
      var values = value.node.value.split(" ");
      return values.length === 1 ? t.memberExpression(cssmodule, t.stringLiteral(values[0]), true) : t.templateLiteral(spacedTemplateElements(values.length + 1), values.map(function (v) {
        return t.memberExpression(cssmodule, t.stringLiteral(v), true);
      }));
    } else if (t.isIdentifier(value)) {
      // TODO: need to validate what type of node this identifier refers to
      return t.memberExpression(cssmodule, value.node, true);
    } else if (t.isMemberExpression(value)) {
      return t.memberExpression(cssmodule, value.node, true);
    } else if (t.isConditionalExpression(value)) {
      replaceConditionalExpression(value.get("consequent"), cssmodule);
      replaceConditionalExpression(value.get("alternate"), cssmodule);
      return value.node;
    } else {
      logOnce("computeClassName", value.node);
      return value.node;
    }
  };

  var isArrayJoin = function isArrayJoin(path) {
    return t.isCallExpression(path.node) && t.isMemberExpression(path.node.callee) && t.isIdentifier(path.node.callee.property, { name: "join" }) && (t.isArrayExpression(path.node.callee.object) || t.isIdentifier(path.node.callee.object) // TODO: resolve identifier
    );
  };

  /**
   * Resolves a path to most conservative Path to be converted. If there is only
   * one reference to an Identifier we can mutate that safely, otherwise we
   * return the original node to be transformed.
   *
   * @param {Path} path an Identifier or ArrayExpression Path
   * @return {Path} the conservatively resolved path
   */
  var conservativeResolve = function conservativeResolve(path) {
    // if it’s not an identifier we already have what we need
    if (!t.isIdentifier(path.node)) {
      return path;
    }
    var binding = path.scope.getBinding(path.node.name);

    // if there is only one reference, we can mutate that directly
    if (binding.references === 1) {
      if (t.isVariableDeclarator(binding.path)) {
        return binding.path.get("init");
      }
      console.warn("TODO: ensure this branch is tracked");
      return binding.path;
    }

    // else we should probably return conservatively only the one we want and
    // transform inline
    return path;
  };

  /**
   * Replaces [...].join(" ") or identifier.join(" ") with the most appropriate
   * cssmodule lookup hash.
   *
   * @param {Path<Identifier|ArrayExpression>} path an Identifier or ArrayExpression Path
   * @param {Node<Identifier>} cssmodule the root identifier to the cssmodules object
   */
  var replaceArrayJoinElements = function replaceArrayJoinElements(path, cssmodule) {
    var arrayExpressionPath = conservativeResolve(path.get("callee").get("object"));
    if (t.isIdentifier(arrayExpressionPath)) {
      arrayExpressionPath.parent.object = t.callExpression(t.memberExpression(arrayExpressionPath.node, t.identifier("map"), false), [t.arrowFunctionExpression([t.identifier("i")], t.memberExpression(cssmodule, t.identifier("i"), true))]);
      return;
    }

    for (var i = 0; i < arrayExpressionPath.node.elements.length; i++) {
      var element = arrayExpressionPath.get("elements", i)[i];
      element.replaceWith(computeClassName(element, cssmodule));
    }
  };

  /**
   * Replaces React.createElement(..., path, ...) with the most appropriate
   * cssmodule lookup hash.
   *
   * @param {Path<Identifier} path an Identifier Path
   * @param {Node<Identifier>} cssmodule the root identifier to the cssmodules object
   */
  var replaceIdentifier = function replaceIdentifier(path, cssmodule) {
    var binding = path.scope.getBinding(path.node.name);

    var updateProperty = function updateProperty(property) {
      if (property.node.key.name === "className") {
        updateClassName(property.get("value"), cssmodule); // eslint-disable-line
      }
    };

    // if there is only one reference, we can mutate that directly
    if (binding.references === 1) {
      if (t.isVariableDeclarator(binding.path)) {
        var sourceNode = binding.path.get("init");
        if (t.isNullLiteral(sourceNode)) {
          return;
        } else if (t.isCallExpression(sourceNode)) {
          sourceNode.get("argument").forEach(function (arg) {
            if (t.isObjectExpression(arg)) {
              arg.get("properties").forEach(updateProperty);
            }
          });
        } else if (t.isObjectExpression(sourceNode)) {
          sourceNode.get("properties").forEach(updateProperty);
        } else if (t.isIdentifier(sourceNode)) {
          replaceIdentifier(sourceNode, cssmodule);
        } else {
          // updateClassName(binding.path.get("init"), cssmodule); // eslint-disable-line
        }
      } else {
          logOnce("replaceIdentifier[maybe]", binding.path);
        }
    } else {
      console.warn("TODO: replaceIdentifier multiple references");
    }
  };

  /**
   * Updates a callExpression value with the most appropriate CSS Module lookup.
   *
   * @param {Path} callExpression <jsx className={value()} />
   * @param {Node<Identifier>} cssmodule cssmodule identifier
   */
  var replaceCallExpression = function replaceCallExpression(callExpression, cssmodule) {
    if (isArrayJoin(callExpression)) {
      return replaceArrayJoinElements(callExpression, cssmodule);
    }

    // this is just mean
    callExpression.replaceWith(t.callExpression(t.memberExpression(t.callExpression(t.memberExpression(t.callExpression(t.memberExpression(callExpression.node, t.identifier("split"), false), [t.stringLiteral(" ")]), t.identifier("map")), [t.arrowFunctionExpression([t.identifier("i")], t.memberExpression(cssmodule, t.identifier("i"), true))]), t.identifier("join"), false), [t.stringLiteral(" ")]));
  };

  var updateClassName = function updateClassName(value, cssmodule) {
    if (t.isStringLiteral(value)) {
      value.replaceWith(computeClassName(value, cssmodule));
    } else if (t.isMemberExpression(value)) {
      value.replaceWith(computeClassName(value, cssmodule));
    } else if (t.isIdentifier(value)) {
      var binding = value.scope.getBinding(value.node.name);
      if (t.isVariableDeclarator(binding.path.node)) {
        console.log("updateClassName with variablie declarator init");
        updateClassName(binding.path.get("init"), cssmodule);
      } else {
        value.replaceWith(computeClassName(value, cssmodule));
      }
    } else if (t.isCallExpression(value)) {
      replaceCallExpression(value, cssmodule);
    } else if (t.isObjectProperty(value)) {
      updateClassName(value.get("value"), cssmodule);
    } else if (t.isConditionalExpression(value)) {
      updateClassName(value.get("consequent"), cssmodule);
      updateClassName(value.get("alternate"), cssmodule);
    } else if (t.isLogicalExpression(value) || t.isBinaryExpression(value)) {
      updateClassName(value.get("left"), cssmodule);
      updateClassName(value.get("right"), cssmodule);
    } else {
      logOnce("updateClassName", value);
    }
  };

  /**
   * Updates a JSX className value with the most appropriate CSS Module lookup.
   *
   * @param {Path} value <jsx className={value} />
   * @param {Node<Identifier>} cssmodule cssmodule identifier
   */
  var updateJSXClassName = function updateJSXClassName(value, cssmodule) {
    if (!t.isJSXExpressionContainer(value)) {
      value.replaceWith({
        type: "JSXExpressionContainer",
        expression: value.node
      });
    }
    updateClassName(value.get("expression"), cssmodule);
  };

  var buildRequire = (0, _babelTemplate2.default)("\n    const IMPORT_NAME = require(SOURCE);\n  ");

  // TODO: template doesn't work for import.
  var buildImport = function buildImport(_ref2) {
    var IMPORT_NAME = _ref2.IMPORT_NAME;
    var SOURCE = _ref2.SOURCE;
    return t.importDeclaration([t.importDefaultSpecifier(IMPORT_NAME)], SOURCE);
  };

  return {
    visitor: {
      JSXAttribute: function JSXAttribute(path, state) {
        if (state[BAIL_OUT]) {
          return;
        }

        if (path.get("name").node.name !== "className") {
          return;
        }

        if (!state.cssModuleId) {
          state.cssModuleId = path.scope.generateUidIdentifier(ROOT_CSSNAMES_IDENTIFIER);
        }

        updateJSXClassName(path.get("value"), state.cssModuleId);
      },
      CallExpression: function CallExpression(path, state) {
        if (state[BAIL_OUT]) {
          return;
        }

        var isCreateElementCall = matchesPatterns(path.get("callee"), ["React.createElement", "_react2.default.createElement"]);

        if (!isCreateElementCall) {
          return;
        }

        if (!state.cssModuleId) {
          state.cssModuleId = path.scope.generateUidIdentifier(ROOT_CSSNAMES_IDENTIFIER);
        }

        var updateProperty = function updateProperty(property) {
          if (property.node.key.name === "className") {
            updateClassName(property.get("value"), state.cssModuleId);
          }
        };

        var argument = path.get("arguments")[1];
        if (t.isNullLiteral(argument)) {
          return;
        } else if (t.isCallExpression(argument)) {
          argument.get("arguments").forEach(function (arg) {
            if (t.isObjectExpression(arg)) {
              arg.get("properties").forEach(updateProperty);
            }
          });
        } else if (t.isObjectExpression(argument)) {
          argument.get("properties").forEach(updateProperty);
        } else if (t.isIdentifier(argument)) {
          replaceIdentifier(argument, state.cssModuleId);
        } else {
          logOnce("CallExpression Visitor", argument.node, path);
        }
      },

      Program: {
        enter: function enter(path, state) {
          if (!state.opts.path || new RegExp(state.opts.path).test(state.file.opts.filename)) {
            // detect if this is likely compiled source
            if (path.scope.getBinding("_interopRequireDefault")) {
              state.transformingOutput = true;
              state.cssModuleId = path.scope.generateUidIdentifier(ROOT_CSSNAMES_IDENTIFIER);
            }
          } else {
            state[BAIL_OUT] = true;
          }
        },
        exit: function exit(path, state) {
          if (state[BAIL_OUT]) {
            return;
          }

          if (state.cssModuleId) {
            var importOpts = {
              IMPORT_NAME: state.cssModuleId,
              SOURCE: t.stringLiteral((0, _path.join)(process.env.NODE_ENV === "test" ? "" : process.cwd(), state.opts.cssmodule))
            };

            var firstChild = path.get("body")[0];
            var leadingComments = firstChild.node.leadingComments;

            delete firstChild.node.leadingComments;

            // currently we’re using the heuristic if the module system is using
            // commonjs then we'll export as commonjs
            path.get("body")[0].insertBefore(state.opts.modules === "commonjs" || state.transformingOutput ? buildRequire(importOpts) : buildImport(importOpts));
            path.get("body")[0].node.leadingComments = leadingComments;
          }
        }
      }
    }
  };
};